#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <sys/ioctl.h>
#include <stdint.h>

#include "wave_header.h"

/*****************************************************************************/
/*  TUNABLE PARAMETERS                                                       */ 
/*****************************************************************************/

/*
 *  The version number.
 */
#define VERSION "0.1"

/*
 *  Default sample rate.  Overridden by a -r option.
 */
#define DEFAULT_RATE 48000

/*
 *  End stop.  Maximum +/- percentage deviation from the nominal delay that
 *  the tracking will be allowed to use.
 */
#define END_STOP 0.5    /* Percent */

/*
 *  Number of bytes to read/write at a time.
 */
#define NREAD 4096

/*
 *  Number of comb delay stages.  Overridden by a -s option.
 */
#define DEFAULT_STAGES 4

/*
 *  Seconds between iterations of the line frequency tracking. Overridden by
 *  a -t option.
 */
#define DEFAULT_TRACKING_CYCLE 1

/*
 *  The default encoding, 8 means unsigned bytes, 16 means signed 16 bit words.
 *  Overridden by -8 and -16 options.
 */
#define DEFAULT_ENCODING 16

/*
 *  This constant sets the nominal slewing rate of the period tracking. This
 *  value gives stable locking under most conditions. The filter may lag when
 *  a sudden shift occurs, if this happens too much,  then raise the value,
 *  eg to 0.02 or 0.04.   If the filter jumps out of lock on bursts of noise,
 *  then reduce the value.
 */
#define SLEW_RATE 0.01

/*
 *  Smoothing factor for the period tracking.  Not very critical.
 */
#define COMB_SMOOTH 0.9999

/*
 *  Maximum tracking width.  The max delay difference between the low and high
 *  measurement periods, and the current delay period, units of samples.
 */
#define MAX_TRACK 10

/*
 *  End of tunable definitions.
 */


/*****************************************************************************/
/*  Globals                                                                  */ 
/*****************************************************************************/

int encoding = DEFAULT_ENCODING;		/* 8 or 16 */
int request_rate = DEFAULT_RATE;		/* The requested sample rate */
int actual_sample_rate;					/* The actual sample rate */

double nominal_delay;
double nominal_frequency;
int auto_tracking_flag;
int wave_file_mode_flag;
int debug_flag;
wave_header *in_header;
int channels;


/*****************************************************************************/
/*  The Comb Filter                                                          */ 
/*****************************************************************************/

short *comb_buffer;                           /* Points to comb delay buffer */
int comb_len;                       /* Total length of the comb delay buffer */
int comb_lp = 0;      /* Load offset into comb_buffer, next location to load */
int comb_mask;                                    /* Mask for buffer offsets */

int comb_state = 1;           /* 1 means filter is in, 0 means filter is out */

int comb_stages = DEFAULT_STAGES;       /* The number of delay stages in use */
float comb_delay;      /* Current comb delay, per stage, in units of samples */
float comb_adj;                      /* Incremental adjustment to comb delay */
float max_delay, min_delay;                  /* Allowed limits to comb delay */

float track_low = 0;                               /* Low period accumulator */
float track_mid = 0;            /* Middle period accumulator - current delay */
float track_high = 0;                             /* High period accumulator */

float tracking_width = 1.0; /* Offset in period space of the tracking guides */
float tracking_cycle = DEFAULT_TRACKING_CYCLE;  /* How often to revise delay */


/*
 *  Initialise the comb filter: allocate buffer memory and set some globals.
 *  The buffer must be long enough for the max possible delay, plus an extra
 *  allowance for the high period tracking.  We then round up to the next
 *  power of two for efficiency.
 */

void setup_comb( void)
{
int i;
comb_delay = nominal_delay * actual_sample_rate / 1000.0;
max_delay = comb_delay * (1 + END_STOP / 100.0);
min_delay = comb_delay * (1 - END_STOP / 100.0);

comb_len = comb_stages * (1 + MAX_TRACK + max_delay);
for(i = 1; i < comb_len; ) i <<= 1;

comb_len = i;
comb_mask = comb_len-1;

if( (comb_buffer = (short *) malloc( sizeof( short) * comb_len)) == NULL)
	{
	fprintf(stderr, "humfilt: setup_comb(): cannot allocate memory for comb_buffer, aborting.\n");

	exit( 1);
	}

} /* end function setup_comb */



/*
 *  Comb summing function.  Sum over the given number of delay stages, each
 *  of length delay_steps samples.  Use linear interpolation between samples.
 *
 *  At high sample rates, adequate results might be obtained without the linear
 *  interpolation, which can save quite a bit of CPU time. 
 */

static inline int comb_sum( int stages, float delay_steps)
{
int ne;
float f = 0, g = delay_steps;

for(ne = 1; ne <= stages; ne++, g += delay_steps)
	{
	int gf = g;
	int p, f1, f2;

	p = comb_lp - gf; 
	f1 = comb_buffer[ p-- & comb_mask]; 
	f2 = comb_buffer[ p & comb_mask]; 
	f += f1 + (f2 - f1) * (g - gf);
	}

return rint(f / stages);
} /* end function comb_sum */


/*
 *  Revise the filter delay.  Compare the mean square amplitudes of the high
 *  and low tracking signals and adjust the delay to bring them nearer to 
 *  equality.  
 */

static void adjust_delay( void)
{
float f;
float balance = track_high/track_low;   /* Equals 1.0 when in tune */
float mean = track_high * track_low/track_mid/track_mid;

if( balance > 1) comb_adj = 1;
else
	{
	comb_adj = -1;
	balance = 1 / balance;
	}

f = pow( balance, 10);  if(f > 10) f = 10;
comb_adj *= f * SLEW_RATE/comb_stages;

if( mean < 0.5) tracking_width /= 2;
else if( mean > 0.8) tracking_width *= 2;

if( tracking_width < 0.1) tracking_width = 0.1;
else if( tracking_width > MAX_TRACK) tracking_width = MAX_TRACK;

/* Revise the delay, and constrain within the end stops. */

if(auto_tracking_flag)
	{
	comb_delay += comb_adj * actual_sample_rate/1000.0;
	}
	
if( comb_delay > max_delay) comb_delay = max_delay;
if( comb_delay < min_delay) comb_delay = min_delay;
} /* end function adjust_delay */



/*
 *  Report status info.
 */

static void print_info( float peak, float sum_square, int count)
{
fprintf( stderr, 
"peak=%.3f rms=%0.3f del=%6.3f %+.4f mS low=%.2f high=%.2f wid=%.2f auto_tracking=%d wave_file_mode=%d\n",
peak/32768,
sqrt( sum_square/count)/32768, 
1000 * comb_delay/actual_sample_rate,
comb_adj,
track_low/track_mid, track_high/track_mid, tracking_width, auto_tracking_flag, wave_file_mode_flag);
} /* end function print_info */


static void inline smooth_accumulate(float *a, float val)
{
*a = *a * COMB_SMOOTH + val * (1-COMB_SMOOTH);
} /* end function smooth_accumulate */


static void inline leaky_integrate(float *a, float val)
{
*a = (*a + val) * 0.999;
} /* end function leaky_integrate */


/*
 *  This function is called with each signal sample.
 */

short comb_filter( short s)
{
static float p_low = 0, p_mid = 0, p_high = 0;
static float sum_square = 0;
static float peak = 0;
static int count = 0;
int t;

/*
*  Insert the incoming sample into the buffer, and keep track of the mean
*  square and peak values.
*/
 
comb_buffer[comb_lp] = s;        /* Store incoming sample in comb buffer */

sum_square += s * s;                /* Accumulate the mean square amplitude */
if( s > peak) peak = s;
else
if( s < -peak) peak = -s;

/*
*  Run the comb summing function three times, for the low, high, and middle
*  periods.  Integrate each sum over time to give a 6db/octave roll-off. 
*  Then smooth the squared amplitudes of the integrated sums.
*/

t = comb_sum( comb_stages, comb_delay - tracking_width);
leaky_integrate( &p_low, t);
smooth_accumulate( &track_low, p_low * p_low);
 
t = comb_sum( comb_stages, comb_delay + tracking_width);
leaky_integrate( &p_high, t);
smooth_accumulate( &track_high, p_high * p_high);

t = comb_sum( comb_stages, comb_delay);
leaky_integrate( &p_mid, t);
smooth_accumulate( &track_mid, p_mid * p_mid);

/*
*  Revise the delay every tracking_cycle secs.
*/

if(++count > actual_sample_rate * tracking_cycle)
	{
	if( track_mid != 0)              /*  Must have a non-zero input signal */
		{
		adjust_delay();
		print_info( peak, sum_square, count);
		}

	sum_square = peak = count = 0;
	}

comb_lp = (comb_lp + 1) & comb_mask;
return comb_state ? s - t : s;
} /* end function comb_filter */


/*****************************************************************************/
/*  Sound Processing                                                         */ 
/*****************************************************************************/

/*
 *  This function copies sound card input to output, passing each sample
 *  through the comb filter.
 */

void process_sound( void)
{
unsigned char buff[ NREAD];
size_t a;

while( 1) 
	{
	int q = 0;
  
	errno = 0; 

	a = fread(buff, sizeof(char), NREAD, stdin);
	if(a == 0) // EOF or error
		{
		return;
		}
		

	

	if( encoding == 8)      /* Samples are 8 bit unsigned */
		{
		for(q = 0; q < NREAD; q++)
			{
			int sample = ((short)buff[q] - 128) << 8;

			sample = comb_filter(sample);
	
			buff[q] = (unsigned char)((sample >> 8) + 128);
			} 
		}
	else                    /* Samples are 16 bit signed */     
		{
		for(q = 0; q < NREAD / 2; q++)
			{
			int sample = *((short *)(buff) + q);

			sample = comb_filter(sample);

			*((short *)(buff) + q) = sample;
			}
		}

	q = 0;
	errno = 0; 

	fwrite(buff, sizeof(char), a, stdout);
	}
} /* end fucntion process_sound */


/*****************************************************************************/
/*  Main Program                                                             */ 
/*****************************************************************************/

void usage( void)
{
fprintf(stderr, "\nHumfilter-%s, based on code by Paul Nicholson, wavefile interface by Jan Panteltje.\n", VERSION);
fprintf(stderr, "Command line options are:\n");
fprintf(stderr, " -8             Use 8 bit unsigned bytes.\n");
fprintf(stderr, " -16            Use 16 bit signed words.\n");
//fprintf(stderr, " -c channels    Audio channels, default 1 (mono).\n");       
fprintf(stderr, " -d             debug, prints functions and arguments\n");
fprintf(stderr, " -f frequency   Use float.\n");
fprintf(stderr, " -k             auto tracking.\n");
fprintf(stderr, " -r rate        Use rate samples/sec.\n");
fprintf(stderr, " -s stages      Use this number of delay stages.\n");
fprintf(stderr, " -t secs        Revise delay every this many seconds.\n");
fprintf(stderr, " -w             wave file mode, first 44 bytes (the header) is copied verbatim.\n");
fprintf(stderr, "\n default is: -%d -f %f -r %d -s %d -t %.1f\n",\
(int) DEFAULT_ENCODING, (float) nominal_frequency, (int) DEFAULT_RATE, (int) DEFAULT_STAGES, (float) DEFAULT_TRACKING_CYCLE);
fprintf(stderr, " nominal delay is: %.3f mS.\n", (float) nominal_delay);
fprintf(stderr, "Examples:\n");
fprintf(stderr, " Play a 48000 samples per second mono wave file with different values for f to find the best f for minimum hum,\n"); 
fprintf(stderr, " bs=44 skips the click from the wave header, you can start later in the file with a higher number for skip.\n"); 
fprintf(stderr, "  dd if=sound_with_hum.wav bs=44 skip=1 | humfilter -r 48000 -16 -s 4 -f 59.998 | sox -r 48000 -w -c 1 -s -t raw - -r 48000 -w -c 1 -s -t ossdsp /dev/dsp\n");
fprintf(stderr, " Create a hum free file with the f value found:\n"); 
fprintf(stderr, "  cat sound_with_hum.wav | humfilter -w -s 4 -f 59.998 > sound_no_hum.wav\n");
}


int main( int argc, char *argv[])
{
unsigned char buffer[44];

/* defaults */
debug_flag = 0;
nominal_frequency = 20.0;
auto_tracking_flag = 0;
wave_file_mode_flag = 0;
channels = 1;
/* end defaults */

if(argc == 1)
	{
	usage();
	exit(1);		
	}

/* Deal with any command line options. */
while( argc > 1)
	{
	if( !strcmp( argv[1], "-8"))
		{
		encoding = 8;
		argc--; argv++;
		}
	else if( !strcmp( argv[1], "-16"))
		{
		encoding = 16;
		argc--; argv++;
		}
/*
	else if( !strcmp( argv[1], "-c") && --argc > 1)
		{
		argv++; channels = atoi(argv[1]); argc--; argv++;
		}
*/
	else if( !strcmp( argv[1], "-d") )
		{
		debug_flag = 1;
		argc--; argv++;
		}
	else if( !strcmp( argv[1], "-f") && --argc > 1)
		{
		argv++; nominal_frequency = atof(argv[1]); argc--; argv++;
		}
	else if( !strcmp( argv[1], "-k") )
		{
		auto_tracking_flag = 1;
		argc--; argv++;
		}
	else if( !strcmp( argv[1], "-r") && --argc > 1)
		{
		argv++; request_rate = atoi( argv[1]); argc--; argv++;
		}
	else if( !strcmp( argv[1], "-s") && --argc > 1)
		{
		argv++; comb_stages = atoi( argv[1]); argc--; argv++;
		}
	else if( !strcmp( argv[1], "-t") && --argc > 1)
		{
		argv++; tracking_cycle = atof( argv[1]); argc--; argv++;
		}
	else if( !strcmp( argv[1], "-w") )
		{
		wave_file_mode_flag = 1;
		argc--; argv++;
		}
	else 
		{
		fprintf( stderr, "humfilt: unrecognised option: [%s]\n", argv[1]);
		usage();
		exit( 1);
		}
	} /* end while process command line arguments */

if(wave_file_mode_flag)
	{
	in_header = (wave_header *) malloc(44);
	if(! in_header)
		{
		fprintf(stderr, "humfilt: main(): could not allocate space for in_header, aborting.\n");
		
		exit(1);
		}

	/* copy wave header */
	fread(buffer, sizeof(char), 44, stdin);
	fwrite(buffer, sizeof(char), 44, stdout);

	/* override some parameters using the wave header info */
	//encoding
	//request_rate

	if( (buffer[0] != 'R') || (buffer[1] != 'I') || (buffer[2] != 'F') || (buffer[3] != 'F') )
		{
		fprintf(stderr, "humfilt: input file is not a RIFF file, aborting.\n");
	
		exit(1);			
		}

	if( (buffer[8] != 'W') || (buffer[9] != 'A') || (buffer[10] != 'V') || (buffer[11] != 'E') )
		{
		fprintf(stderr, "humfilt: input file is not a WAVE file, aborting.\n");
	
		exit(1);			
		}

	if( (buffer[12] != 'f') || (buffer[13] != 'm') || (buffer[14] != 't') || (buffer[15] != ' ') )
		{
		fprintf(stderr, "humfilt: WARNING input file header does not contain 'fmt '.\n");
	
//		exit(1);			
		}

	in_header -> length =		buffer[ 4] | (buffer[ 5] << 8) | (buffer[6] << 16) | (buffer[7] << 24); 
	in_header -> length_chunk =	buffer[16] | (buffer[17] << 8) | (buffer[18] << 16) | (buffer[19] << 24);
	in_header -> format =		buffer[20] | (buffer[21] << 8);
	in_header -> modus =		buffer[22] | (buffer[23] << 8);
	in_header -> sample_fq =	buffer[24] | (buffer[25] << 8) | (buffer[26] << 16) | (buffer[27] << 24);
	in_header -> byte_p_sec =	buffer[28] | (buffer[29] << 8) | (buffer[30] << 16) | (buffer[31] << 24);
	in_header -> byte_p_spl =	buffer[32] | (buffer[33] << 8);
	in_header -> bit_p_spl =	buffer[34] | (buffer[35] << 8);
	in_header -> data_length =	buffer[40] | (buffer[41] << 8) | (buffer[42] << 16) | (buffer[43] << 24);

	if(debug_flag)
		{
		fprintf(stderr, "in_header -> length=%d\n", in_header -> length);
		fprintf(stderr, "in_header -> length_chunk=%d\n", in_header -> length_chunk);
		fprintf(stderr, "in_header -> format=%d\n", in_header -> format);
		fprintf(stderr, "in_header -> modus=%d\n", in_header -> modus);
		fprintf(stderr, "in_header -> sample_fq=%d\n", in_header -> sample_fq);
		fprintf(stderr, "in_header -> byte_p_sec=%d\n", in_header -> byte_p_sec);
		fprintf(stderr, "in_header -> byte_p_spl=%d\n", in_header -> byte_p_spl);
		fprintf(stderr, "in_header -> bit_p_spl=%d\n", in_header -> bit_p_spl);
		fprintf(stderr, "in_header -> data_length=%d\n", in_header -> data_length);
		fprintf(stderr, "\n");
		}

	if(in_header -> modus != 1)
		{
		fprintf(stderr, "humfilt: input file is not a mono file, aborting.\n");
	
		exit(1);			
		}

	if(in_header -> length_chunk != 16)
		{
		fprintf(stderr, "humfilt: input file incorrect header, length_chunk != 16, aborting.\n");
	
		exit(1);			
		}

	if(in_header -> format != 1)
		{
		fprintf(stderr, "humfilt: input file is not a PCM file, aborting.\n");
	
		exit(1);			
		}
	
	encoding = in_header -> bit_p_spl;
	request_rate = in_header -> sample_fq;
	channels = in_header -> modus;
	} /* end if wave_file_mode_flag */

nominal_delay = 1000.0 / nominal_frequency;
actual_sample_rate = request_rate;

/* Initialise everything. */
setup_comb();

fprintf( stderr, "%d bit encoding, %d samples/sec\n", encoding, actual_sample_rate);
fprintf( stderr, "%d stages\n", comb_stages);

/* Run the filter. */
process_sound();

exit(0);
} /* end function main */


